import os, sys
import tempfile
import pprint
import logging

verbose = True

def D(s):
    logging.debug("[vdiskman] %s" % s)

def traildigits(strg):
    """
    Return the trailing digits, used to split the partition number off
    """
    idx = len(strg)-1
    while strg[idx].isdigit():
        if len == 0:
            return strg
        idx -= 1
    return strg[idx+1:] 

def freeLoopDev():
    """
    Finds a free loop device
    """
    loops = []
    fd = os.popen("losetup -a")
    for ln in fd.readlines():
        loops.append(ln.split(':')[0])
    for nr in range(0,256):
        if "/dev/loop%i" % nr not in loops:
            return "/dev/loop%i" % nr
    return None

def findPart(dev):
    """
    Find device dev in list of partitions
    """
    if len(dev) > 5 and dev[:5] == "/dev/":
        dev = dev[5:]
    for part in partitions:
        if dev == part.domname:
            return part
    return None

class Wholedisk:
    """
    A whole disk that has partitions
    """
    def __init__(self, physdev, loopfile = None):
        self.partitions = []
        self.domname  = "ovsvdiskpart"
        self.physdev  = physdev
        self.loopfile = loopfile
        self.mapped   = 0
        self.pcount   = self.scanpartitions()
        self.rootpartition = None
        self.fstab    = None

    def loopsetup(self):
        """
        Setup the loop mapping
        """
        if self.loopfile and not self.physdev:
            ldev = freeLoopDev()
            if not ldev:
                raise RuntimeError("No free loop device found")
            if verbose:
                D("losetup %s %s" % (ldev, self.loopfile))
            fd = os.popen("losetup %s %s" % (ldev, self.loopfile))
            if fd.close():
                raise RuntimeError("Failure setting up loop dev")
            self.physdev = ldev

    def loopclean(self):
        """
        Delete the loop mapping
        """
        if self.loopfile and self.physdev:
            if verbose:
                D("losetup -d %s" % self.physdev)
            fd = os.popen("losetup -d %s" % self.physdev)
            self.physdev = None
            return fd.close()
    
    def scanpartitions(self):
        """
        Scan device for partitions (kpartx -l) and set up data structures,
           Returns number of partitions found.
        """
        self.loopsetup()
        if not self.physdev:
            raise RuntimeError("No physical device? %s" % self.__repr__())
        fd = os.popen("kpartx -l %s" % self.physdev)
        pcount = 0
        for line in fd.readlines():
            line = line.strip()
            (pname, params) = line.split(':')
            pno = int(traildigits(pname.strip()))
            pname = "/dev/mapper/" + pname
            self.partitions.append(Partition(self, self.domname + '%i' % pno, pname))
            pcount += 1
        fd.close()
        if not pcount:
            if self.loopfile:
                ref = self
            else:
                ref = None
            self.partitions.append(Partition(ref, self.domname, self.physdev))
        self.loopclean()
        return pcount

    def activatepartitions(self):
        """
        Set up loop mapping and device-mapper mappings
        """
        if not self.mapped:
            self.loopsetup()
            if self.pcount:
                if verbose:
                    D("kpartx -a %s" % self.physdev)
                fd = os.popen("kpartx -a %s" % self.physdev)
                fd.close()
        self.mapped += 1
    
    def deactivatepartitions(self):
        """
        Remove device-mapper mappings and loop mapping
        """
        if not self.mapped:
            return
        self.mapped -= 1
        if not self.mapped:
            if self.pcount:
                if verbose:
                    D("kpartx -d %s" % self.physdev)
                fd = os.popen("kpartx -d %s" % self.physdev)
                fd.close()
            self.loopclean()

    def close(self):
        for part in self.partitions:
            part.umount()
        self.deactivatepartitions()
        self.loopclean()

    def __del__(self):
        if verbose:
            D("Wholedisk d'tor")
        self.close()

    def __repr__(self):
        strg = "[" + self.domname + ","
        if self.physdev:
            strg += self.physdev
        strg += ","
        if self.loopfile:
            strg += self.loopfile
        strg += "," + str(self.pcount) + ",mapped %ix]" % self.mapped
        return strg

    def guessRootPartition(self):
        """Guess root partition.
            If there is a file /etc/fstab on a partition, we assume it's
            the root partition.
            If /etc/fstab is found on multiple partitions, raise a run time
            exception.

            @return root partition
        """
        rootPartition = None
        for part in self.partitions:
            try:
                part.mount()
                # if os.access(part.mountpoint + '/etc/fstab', os.F_OK):
                fstab = os.path.join(part.mountpoint, 'etc/fstab')
                D(fstab)
                if os.path.isfile(fstab):
                    if rootPartition != None:
                        if verbose:
                            D("Found root partition: %s" % part)
                        # part.umount()
                        raise RuntimeError("Found /etc/fstab in more than one partition.")
                    else:
                        rootPartition = part
            except RuntimeError, e:
                D(e)
            # part.umount()

        if rootPartition == None:
            # raise RuntimeError("Didn't find /etc/fstab in any partition.")
            D("Didn't find /etc/fstab in any partition.")

        return rootPartition

    def getPartitionByNumber(self, number):
        if verbose:
            D("partition number: %s" % number)
        import re
        for part in self.partitions:
            m = re.search('.+(\d+)$', part.devname)
            if m.group(0) == number:
                return part

    def getPartitionByLabel(self, label):
        if verbose:
            D("partition label: %s" % label)
        for part in self.partitions:
            if verbose:
                D("volume name: %s" % part.volumename)
            if label == part.getVolumeName():
                return part

    def getPartitionByUuid(self, uuid):
        if verbose:
            D("partitions uuid: %s" % uuid)
        for part in self.partitions:
            if uuid == part.getFsUuid():
                return part

    def getPartitionByDevname(self, devname):
        """
        Get partiton number from device name.
        /dev/sda2
        LABLE=xxx
        UUID=xxx
        ID=xxx
        """
        if devname.startswith('/dev/'):
            import re
            m = re.search('.+(\d+)$', devname)
            return self.getPartitionByNumber(m.group(0))
        elif devname.startswith('LABEL='):
            label = devname[len('LABEL='):]
            return self.getPartitionByLabel(label)
        elif devname.startswith('UUID='):
            uuid = devname[len['UUID=']:]
            return self.getPartitionByUuid(uuid)
        elif devname.startswith('ID='):
            D("device ID= doesn't support yet.")
        elif devname.startswith('NAME='):
            D("device NAME= doesn't support yet.")

    def getPartitionByFilename(self, path):
        """
        A quick method to get partition from file path.
        """
        if self.rootpartition == None:
            self.rootpartition = self.guessRootPartition()
        if self.rootpartition != None:
            if self.fstab == None:
                self.fstab = Fstab(os.path.join(self.rootpartition.mountpoint, 'etc/fstab'))
            devname, mountpoint = self.fstab.find(path)
            return self.getPartitionByDevname(devname)
        else:
            D("Use first partition as root.")
            return self.partitions[0]

    def getMountpathByFilename(self, path):
        """
        A quick method to get mount path.
        """
        if self.rootpartition == None:
            self.rootpartition = self.guessRootPartition()
        if self.rootpartition != None:
            if self.fstab == None:
                self.fstab = Fstab(os.path.join(self.rootpartition.mountpoint, 'etc/fstab'))
            devname, mountpoint = self.fstab.find(path)
            if verbose:
                D("%s on %s,%s" % (path, devname, mountpoint))
            part = self.getPartitionByDevname(devname)
            if verbose:
                D("%s on %s" % (path, part))
                D("join %s %s" % (part.mountpoint, path[len(mountpoint):]))
            return os.path.join(part.mountpoint, path[len(mountpoint):].lstrip('/'))
        else:
            D("User the first partiton as default partition.")
            part = self.partitions[0]
            D("%s on %s" % (path, part))
            return os.path.join(part.mountpoint, path.lstrip('/'))


class Partition:
    def __init__(self, whole = None, domname = None, 
                 physdev = None):
        self.wholedisk  = whole
        self.domname    = domname
        self.physdev    = physdev
        self.mountpoint = None
        self.volumename = None
        self.fsuuid     = None
        self.scanned    = False

    def __del__(self):
        if self.mountpoint:
            self.umount()

    def __repr__(self):
        strg = "[" + self.domname + "," + self.physdev + ","
        if self.mountpoint:
            strg += "mounted on " + self.mountpoint + ","
        else:
            strg += "not mounted,"
        if self.wholedisk:
            return strg + self.wholedisk.__repr__() + "]"
        else:
            return strg + "]"

    def getVolumeName(self):
        if self.scanned == False:
            self.getVolumeNameAndUuid()
        return self.volumename

    def getFsUuid(self):
        if self.scanned == False:
            self.getVolumeNameAndUuid()
        return self.fsuuid

    def getVolumeNameAndUuid(self):
        if self.scanned:
            return
        self.scanned = True
        if verbose:
            D("tune2fs -l %s" % self.physdev)
        fd = os.popen("tune2fs -l %s" % self.physdev)
        for ln in fd.readlines():
            if ln.startswith("Filesystem volume name:"):
                self.volumename = ln[len("Filesystem volume name:"):].strip()
            elif ln.startswith("Filesystem UUID:"):
                self.fsuuid = ln[len("Filesystem UUID:"):].strip()

    def mount(self, fstype = None, options = "rw"):
        if self.mountpoint:
            return
        if self.wholedisk:
            self.wholedisk.activatepartitions()
        mtpt = tempfile.mkdtemp(prefix = "%s." % self.domname)
        mopts = ""
        if fstype:
            mopts += " -t %s" % fstype
        mopts += " -o %s" % options
        if verbose:
            D("mount %s %s %s" % (mopts, self.physdev, mtpt))
        fd = os.popen("mount %s %s %s" % (mopts, self.physdev, mtpt))
        err = fd.close()
        if err:
            raise RuntimeError("Error %i from mount %s %s on %s" % \
                (err, mopts, self.physdev, mtpt))
        self.mountpoint = mtpt
        
    def umount(self):
        if not self.mountpoint:
            return
        if verbose:
            D("umount %s" % self.mountpoint)
        fd = os.popen("umount %s" % self.mountpoint)
        err = fd.close()
        os.rmdir(self.mountpoint)
        if err:
            raise RuntimeError("Error %i from umount %s" % \
                (err, self.mountpoint))
        self.mountpoint = None
        if self.wholedisk:
            self.wholedisk.deactivatepartitions()

class Fstab:
    class FstabEntry:
        def __init__(self, line):
            spline = line.split()
            self.dev, self.mtpt, self.fstype, self.opts = \
                spline[0], spline[1], spline[2], spline[3]
            if len(self.mtpt) > 1:
                self.mtpt = self.mtpt.rstrip('/')
    
    def __init__(self, filename):
        self.entries = []
        fd = open(filename)
        for line in fd.readlines():
            line = line.strip()
            if len(line) == 0 or line[0] == '#':
                continue
            self.entries.append(Fstab.FstabEntry(line))

    def find(self, fname):
        matchlen = 0
        match = None
        fnmlst = fname.split('/')
        for fs in self.entries:
            entlst = fs.mtpt.split('/')
            if entlst == ['','']:
                entlst = ['']
            entln = len(entlst)
            if len(fnmlst) >= entln and fnmlst[:entln] == entlst \
                and entln > matchlen:
                match = fs
                matchlen = entln
        if not match:
            return (None, None)
        return (match.dev, match.mtpt)


def test(vdisk, file):
    def dumpFile(filename):
        try:
            f = open(filename, 'r')
            for ln in f.readlines():
                pprint.pprint('%s' % ln)
            f.close()
        except Exception, e:
            D(e)

    def appendFile(filename, lines):
        try:
            f = open(filename, 'a')
            f.write(lines)
            f.close()
        except Exception, e:
            D(e)

    wdisk = Wholedisk(None, vdisk)
    # D(partitions)
    try:
        """
        rootpartition = wdisk.guessRootPartition()
        fstab = Fstab(os.path.join(rootpartition.mountpoint, 'etc/fstab'))
        path = '/etc/sysconfig/test'
        devname, mountpoint = fstab.find(path)
        part = wdisk.getPartitionByDevname(devname)
        D("%s on %s" % (path, part))
        # dump the file
        realpath = os.path.join(part.mountpoint, path[len(mountpoint):])
        dumpFile(realpath)
        """

        pprint.pprint(file)
        realpath = wdisk.getMountpathByFilename(file)
        dumpFile(realpath)
        """
        import time
        appendFile(realpath, "%s: this is a test.\n" % time.asctime())
        dumpFile(realpath)
        """

        """
        path = '/boot/grub/menu.lst'
        devname, mountpoint = fstab.find(path)
        D("%s on %s" % (path, wdisk.getPartitionByDevname(devname)))
        path = '/etc/sysconfig/oraclevm-template'
        D("%s on %s" % (path, fstab.find(path)))
        """
    except Exception, e:
        D("%s, %s" % (type(e), e.args))
    wdisk.close()

def serialized_call(func):
    func.serialized = True
    return func
    """
    from OVSConstants import CONFIG_DIR
    from OVSFileLock import FileLock
    lock_base = join(CONFIG_DIR, 'lock')
    
    if not exists(lock_base):
        os.makedirs(lock_base)

    lock_file = join(lock_base, "%s.lock" % func.func_name)
    lock = FileLock(lock_file)
    lock.acquire()
    """

if __name__ == '__main__':
    test(sys.argv[1], sys.argv[2])

